/*
 * cntlr_tlv_qos.c - tlv building functions for QoS
 *
 * Copyright (C) 2023 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: maxim.menshikov@iopsys.eu
 *
 */
#if (EASYMESH_VERSION >= 3)

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>
#include <pthread.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>

#include <easy/easy.h>
#include <i1905_wsc.h>

#include <timer_impl.h>
#include <cmdu.h>
#include <1905_tlvs.h>
#include <easymesh.h>
#include <map_module.h>

#include <wifidefs.h>
#include "wifi_dataelements.h"

#include <uci.h>

#include "utils/utils.h"
#include "utils/debug.h"
#include "config.h"
#include "dpp.h"
#include "cntlr.h"
#include "cntlr_ubus.h"

#include "cntlr_map_debug.h"
#include "cntlr_tlv.h"

/* WiFi EasyMesh Specification 17.2.70 - Service Prioritization Rule TLV */
int cntlr_gen_spr_tlv(struct controller *c,
                      struct cmdu_buff *frm,
                      uint32_t rule_id,
                      bool add,
                      uint8_t precedence,
                      uint8_t output,
                      bool always_match)
{
	struct tlv *t;
	struct tlv_spr *tlv_data;
	const uint16_t max_tlv_len = sizeof(struct tlv_spr);

	dbg("%s: --->\n", __func__);
	t = cmdu_reserve_tlv(frm, max_tlv_len);
	if (!t)
		return -1;

	t->type = MAP_TLV_SERVICE_PRIORITIZATION_RULE;
	t->len = sizeof(struct tlv_spr);

	tlv_data = (struct tlv_spr *)t->data;
	BUF_PUT_BE32(tlv_data->rule_id, rule_id);
	tlv_data->add_remove = add;
	tlv_data->precedence = precedence;
	tlv_data->output = output;
	tlv_data->always_match = always_match;

	if (cmdu_put_tlv(frm, t)) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

/* WiFi EasyMesh Specification 17.2.71 - DSCP Mapping Table TLV */
int cntlr_gen_dscp_mapping_table_tlv(struct controller *c,
                                     struct cmdu_buff *frm,
                                     uint8_t dscp_pcp[64])
{
	struct tlv *t;
	struct tlv_dscp_pcp *tlv_data;
	const uint16_t max_tlv_len = sizeof(struct tlv_dscp_pcp);

	dbg("%s: --->\n", __func__);
	t = cmdu_reserve_tlv(frm, max_tlv_len);
	if (!t)
		return -1;

	t->type = MAP_TLV_DSCP_MAPPING_TABLE;
	t->len = sizeof(*tlv_data);

	tlv_data = (struct tlv_dscp_pcp *)t->data;
	memcpy(tlv_data->dscp_pcp, dscp_pcp, sizeof(tlv_data->dscp_pcp));

	if (cmdu_put_tlv(frm, t)) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

#if (EASYMESH_VERSION >= 4)

/* WiFi EasyMesh Specification 17.2.92 - QoS Management Policy TLV */
static int _cntlr_gen_qos_management_policy_tlv(struct controller *c,
                                                struct cmdu_buff *frm,
                                                uint8_t num_mscs_disallowed,
                                                uint8_t **mscs_disallowed,
                                                uint8_t num_scs_disallowed,
                                                uint8_t **scs_disallowed)
{
	struct tlv *t;
	struct sta_macaddr *lst;
	const uint16_t max_tlv_len =
		sizeof(uint8_t) +
		sizeof(uint8_t) * 6 * num_mscs_disallowed +
		sizeof(uint8_t) +
		sizeof(uint8_t) * 6 * num_scs_disallowed +
		sizeof(uint8_t) * 20 /* reserved */;
	size_t i;

	dbg("%s: --->\n", __func__);
	t = cmdu_reserve_tlv(frm, max_tlv_len);
	if (!t)
		return -1;

	t->type = MAP_TLV_QOS_MANAGEMENT_POLICY;
	t->len = max_tlv_len;

	lst = (struct sta_macaddr *)t->data;
	lst->num = num_mscs_disallowed;
	for (i = 0; i < num_mscs_disallowed; ++i) {
		memcpy(lst->macaddr[i], mscs_disallowed[i], sizeof(uint8_t) * 6);
	}

	lst = (struct sta_macaddr *)(t->data +
	    sizeof(uint8_t) /* skip numMSCS */ +
	    sizeof(uint8_t) * 6 *
	    num_mscs_disallowed /* skip MSCS disallowed */);
	lst->num = num_scs_disallowed;
	for (i = 0; i < num_scs_disallowed; ++i) {
		memcpy(lst->macaddr[i], scs_disallowed[i], sizeof(uint8_t) * 6);
	}
	memcpy(lst->macaddr, scs_disallowed,
		   sizeof(uint8_t) * 6 * num_scs_disallowed);

	if (cmdu_put_tlv(frm, t)) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int cntlr_gen_qos_management_policy_tlv(struct controller *c,
                                        struct cmdu_buff *frm)
{
	struct qos_mgmt_policy_elem *elem;
	uint8_t **mscs_data = NULL;
	uint32_t mscs_num = 0;
	uint8_t **scs_data = NULL;
	uint32_t scs_num = 0;
	size_t i;
	int ret = -1;

	/* Count MSCS disallowed/SCS disallowed */
	list_for_each_entry(elem, &c->cfg.qos.policy.mscs_disallowed, list) {
		mscs_num++;
	}

	list_for_each_entry(elem, &c->cfg.qos.policy.scs_disallowed, list) {
		scs_num++;
	}

	/* Fulfill MSCS disallowed/SCS disallowed MAC lists */
	mscs_data = alloca(mscs_num * sizeof(uint8_t *));
	if (mscs_data == NULL) {
		goto out;
	}

	scs_data = alloca(scs_num * sizeof(uint8_t *));
	if (scs_data == NULL) {
		goto out;
	}

	i = 0;
	list_for_each_entry(elem, &c->cfg.qos.policy.mscs_disallowed, list) {
		mscs_data[i] = alloca(6 * sizeof(uint8_t));
		if (mscs_data[i] == NULL)
			goto out;
		memcpy(mscs_data[i], elem->mac, sizeof(elem->mac));
		i++;
	}

	i = 0;
	list_for_each_entry(elem, &c->cfg.qos.policy.scs_disallowed, list) {
		scs_data[i] = alloca(6 * sizeof(uint8_t));
		if (scs_data[i] == NULL)
			goto out;
		memcpy(scs_data[i], elem->mac, sizeof(elem->mac));
		i++;
	}

	ret = _cntlr_gen_qos_management_policy_tlv(c, frm,
	    mscs_num, mscs_data, scs_num, scs_data);

out:
	return ret;
}

/* WiFi EasyMesh Specification 17.2.93 - QoS Management Descriptor TLV */
int cntlr_gen_qos_management_desc_tlv(struct controller *c,
                                      struct cmdu_buff *frm,
                                      uint16_t qmid,
                                      uint8_t  bssid[6],
                                      uint8_t  sta_mac[6],
                                      uint8_t  desc[],
                                      uint32_t desc_size)
{
	struct tlv *t;
	struct tlv_qos_mgmt_desc *tlv_data;
	const uint16_t max_tlv_len = sizeof(struct tlv_qos_mgmt_desc) + desc_size;

	dbg("%s: --->\n", __func__);
	t = cmdu_reserve_tlv(frm, max_tlv_len);
	if (!t)
		return -1;

	t->type = MAP_TLV_QOS_MANAGEMENT_DESCRIPTOR;
	t->len = max_tlv_len;

	tlv_data = (struct tlv_qos_mgmt_desc *)t->data;
	BUF_PUT_BE32(tlv_data->qmid, qmid);
	memcpy(tlv_data->bssid, bssid, sizeof(tlv_data->bssid));
	memcpy(tlv_data->sta, sta_mac, sizeof(tlv_data->sta));
	memcpy(tlv_data->desc, desc, desc_size);

	if (cmdu_put_tlv(frm, t)) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

uint32_t tclas_elem_frame_len(struct tclas_frame_classifier *cls,
                              size_t *err_count)
{
	switch (cls->type)
	{
		case TCLAS_FRAME_CLASSIFIER_TYPE_IP:
		{
			struct tclas_type4 *type4 =
				(struct tclas_type4 *)&cls->params4;

			switch (type4->ip_version)
			{
				case 4:
				{
					return sizeof(cls->type) +
						   sizeof(type4->classifier_mask) +
						   sizeof(type4->ip_version) +
						   sizeof(type4->ipv4);
				}
				case 6:
				{
					return sizeof(cls->type) +
						   sizeof(type4->classifier_mask) +
						   sizeof(type4->ip_version) +
						   sizeof(type4->ipv6);
				}
				default:
				{
					(*err_count)++;
					return 0;
				}
			}
		}
		default:
		{
			(*err_count)++;
			return 0;
		}
	}
}

uint32_t tclas_elem_len(struct tclas_elem *cls, size_t *err_count)
{
	return sizeof(cls->user_priority) +
	       tclas_elem_frame_len(&cls->classifier, err_count);
}

int cntlr_gen_qos_tclas_frame_classifier_elem(
    struct tclas_frame_classifier *cls,
    struct tclas_frame_classifier *in)
{
	cls->type = in->type;

	switch (in->type)
	{
		case TCLAS_FRAME_CLASSIFIER_TYPE_IP:
		{
			struct tclas_type4 *in_type4 =
				(struct tclas_type4 *)
					&in->params4;
			struct tclas_type4 *out_type4 =
				(struct tclas_type4 *)
					&cls->params4;

			out_type4->classifier_mask = in_type4->classifier_mask;
			out_type4->ip_version = in_type4->ip_version;
			switch (in_type4->ip_version)
			{
				case 4:
				{
					memcpy(&out_type4->ipv4,
						   &in_type4->ipv4,
						   sizeof(struct tclas_type4_ipv4));
					BUF_PUT_BE32(out_type4->ipv4.src_ip,
					             in_type4->ipv4.src_ip);
					BUF_PUT_BE32(out_type4->ipv4.dst_ip, in_type4->ipv4.dst_ip);
					BUF_PUT_BE16(out_type4->ipv4.src_port,
					             in_type4->ipv4.src_port);
					BUF_PUT_BE16(out_type4->ipv4.dst_port,
					             in_type4->ipv4.dst_port);
					return 0;
				}
				case 6:
				{
					memcpy(&out_type4->ipv6,
						   &in_type4->ipv6,
						   sizeof(struct tclas_type4_ipv6));
					BUF_PUT_BE16(out_type4->ipv6.src_port,
					             in_type4->ipv6.src_port);
					BUF_PUT_BE16(out_type4->ipv6.dst_port,
					             in_type4->ipv6.dst_port);
					return 0;
				}
				default:
				{
					break;
				}
			}
			break;
		}
		default:
		{
			break;
		}
	}

	return -1;
}

int cntlr_gen_qos_tclas_elem(uint8_t *data,
                             struct tclas_elem *in_elem,
                             uint32_t elem_len)
{
	struct ie *ieee_elem = (struct ie *)data;
	struct tclas_elem *out_elem = (struct tclas_elem *)
		ieee_elem->data;

	ieee_elem->eid = IE_TCLAS;
	ieee_elem->len = elem_len;

	out_elem->user_priority = in_elem->user_priority;

	return cntlr_gen_qos_tclas_frame_classifier_elem(&out_elem->classifier,
	                                                 &in_elem->classifier);
}

/*
 * Part 11: Wireless LAN Medium Access Control (MAC) and Physical Layer (PHY)
 * Specifications 9.4.2.121 SCS Descriptor element
 */
uint8_t *cntlr_gen_qos_scs_desc(struct controller *c,
                                uint8_t request_type,
                                struct scs_desc_usr *in,
                                uint32_t *out_len)
{
	uint8_t  desc_len;
	uint8_t *desc;
	struct ie *elem;
	struct scs_desc *scs_desc;
	size_t err_count = 0;

	desc_len = sizeof(scs_desc->scsid) + sizeof(scs_desc->request_type);

	switch (request_type)
	{
		case SCS_DESC_REQ_TYPE_ADD:
		case SCS_DESC_REQ_TYPE_CHANGE:
		{
			size_t i;

			desc_len += sizeof(uint8_t) * 3;/* intra_access_category_priority */

			for (i = 0; i < in->tclas_elems_count; ++i) {
				desc_len += tclas_elem_len(&in->tclas_elems[i], &err_count);
			}

			if (in->processing_present) {
				desc_len += sizeof(struct ie);
				desc_len += sizeof(uint8_t);
			}
			break;
		}
		case SCS_DESC_REQ_TYPE_REMOVE:
		default:
		{
			break;
		}
	}

	if (err_count != 0)
		return NULL;

	desc = calloc(sizeof(struct ie) + desc_len, sizeof(uint8_t));
	if (desc == NULL)
		return NULL;

	elem = (struct ie *)desc;

	elem->eid = IE_SCS_DESC;
	elem->len = desc_len;

	scs_desc = (struct scs_desc *)elem->data;
	scs_desc->scsid = in->scsid;
	scs_desc->request_type = request_type;

	switch (request_type)
	{
		case SCS_DESC_REQ_TYPE_ADD:
		case SCS_DESC_REQ_TYPE_CHANGE:
		{
			struct ie *proc_elem;
			uint8_t *tmp = (uint8_t *)scs_desc->add_change.hdr.tclas_elems;
			size_t   i;
			uint32_t prio;

			BUF_PUT_BE32(
				prio,
				in->intra_access_category_priority);
			scs_desc->add_change.hdr.intra_access_category_priority = prio;
			for (i = 0; i < in->tclas_elems_count; ++i) {
				uint32_t tmp_len = tclas_elem_len(&in->tclas_elems[i],
				                                  &err_count);

				cntlr_gen_qos_tclas_elem(tmp, &in->tclas_elems[i], tmp_len);
				tmp += tmp_len;
			}

			if (in->processing_present) {
				/* After all elements */
				proc_elem = (struct ie *)tmp;
				proc_elem->eid = IE_TCLAS_PROCESSING;
				proc_elem->len = sizeof(uint8_t);
				proc_elem->data[0] = in->processing;
			}
			break;
		}
		case SCS_DESC_REQ_TYPE_REMOVE:
		default:
		{
			break;
		}
	}

	*out_len = sizeof(struct ie) + desc_len;

	return desc;
}

/*
 * Part 11: Wireless LAN Medium Access Control (MAC) and Physical Layer (PHY)
 * Specifications 9.4.2.243 SCS Descriptor element
 */
uint8_t *cntlr_gen_qos_mscs_desc(struct controller *c,
                                 uint8_t request_type,
                                 struct mscs_desc_usr *in,
                                 uint32_t *out_len)
{
	uint8_t  desc_len;
	uint8_t *desc;
	struct ext_ie *elem;
	struct mscs_desc *mscs_desc;
	size_t err_count = 0;

	desc_len = sizeof(mscs_desc->request_type) +
	           sizeof(mscs_desc->up_bitmap) +
	           sizeof(mscs_desc->up_limit) +
	           sizeof(mscs_desc->stream_timeout);

	switch (request_type)
	{
		case SCS_DESC_REQ_TYPE_ADD:
		case SCS_DESC_REQ_TYPE_CHANGE:
		{
			size_t i;

			for (i = 0; i < in->tclas_elems_count; ++i) {
				desc_len += tclas_elem_len(&in->tclas_elems[i], &err_count);
			}

			break;
		}
		case SCS_DESC_REQ_TYPE_REMOVE:
		default:
		{
			break;
		}
	}

	if (err_count != 0)
		return NULL;

	desc = calloc(sizeof(struct ext_ie) + desc_len,
				  sizeof(uint8_t));
	if (desc == NULL)
		return NULL;

	elem = (struct ext_ie *)desc;

	elem->eid = IE_EXT;
	elem->eid_ext = IE_EXT_MSCS_DESC;
	elem->len = desc_len;

	mscs_desc = (struct mscs_desc *)elem->data;
	mscs_desc->request_type = request_type;
	mscs_desc->up_bitmap = in->up_bitmap;
	mscs_desc->up_limit = in->up_limit;
	BUF_PUT_BE32(mscs_desc->stream_timeout, in->stream_timeout);

	switch (request_type)
	{
		case SCS_DESC_REQ_TYPE_ADD:
		case SCS_DESC_REQ_TYPE_CHANGE:
		{
			uint8_t *tmp = (uint8_t *)mscs_desc->tclas_elems;
			size_t   i;

			for (i = 0; i < in->tclas_elems_count; ++i) {
				uint32_t tmp_len = tclas_elem_len(&in->tclas_elems[i],
				                                  &err_count);

				cntlr_gen_qos_tclas_elem(tmp, &in->tclas_elems[i], tmp_len);
				tmp += tmp_len;
			}

			break;
		}
		case SCS_DESC_REQ_TYPE_REMOVE:
		default:
		{
			break;
		}
	}

	*out_len = sizeof(struct ext_ie) + desc_len;

	return desc;
}

/*
 * WiFi QoS Management Specification Version 2.0
 * 5.4 QoS Management attribute
 */
uint32_t cntlr_gen_qos_mgmt_attr_desc_len(struct controller *c,
                                          struct qos_mgmt_attr_usr *in,
                                          size_t *err_count)
{
	uint32_t desc_len;

	desc_len = sizeof(struct ie);

	switch (in->attribute_id)
	{
		case QOS_MGMT_ATTR_ID_PORT_RANGE:
		{
			desc_len += sizeof(in->port_range);
			break;
		}
		case QOS_MGMT_ATTR_ID_DSCP_POLICY:
		{
			desc_len += sizeof(in->dscp_policy);
			break;
		}
		case QOS_MGMT_ATTR_ID_TCLAS:
		{
			desc_len += tclas_elem_frame_len(&in->tclas.classifier, err_count);
			break;
		}
		case QOS_MGMT_ATTR_ID_DOMAIN_NAME:
		{
			desc_len += in->domain_name.domain_name_len;
			break;
		}
		default:
		{
			(*err_count)++;
			break;
		}
	}

	return desc_len;
}

/*
 * WiFi QoS Management Specification Version 2.0
 * 5.3 QoS Management element
 */
uint8_t *cntlr_gen_qos_mgmt_elem_desc(struct controller *c,
                                      uint8_t request_type,
                                      struct qos_mgmt_attr_usr *in,
                                      uint32_t *out_len)
{
	uint8_t *data;
	uint32_t desc_len;
	uint32_t inner_len;
	uint8_t oui[3] = WFA_OUI;
	struct qos_mgmt_ie *ve;
	struct qos_mgmt_attr_usr *attr;
	size_t err_count = 0;

	inner_len = cntlr_gen_qos_mgmt_attr_desc_len(c, in, &err_count);
	if (err_count != 0) {
		return NULL;
	}

	desc_len = sizeof(struct qos_mgmt_ie) +
			   inner_len;

	data = calloc(desc_len, sizeof(uint8_t));
	if (data == NULL)
		return NULL;

	ve = (struct qos_mgmt_ie *)data;
	ve->hdr.eid = IE_VEND_SPEC;
	ve->hdr.len = inner_len;
	memcpy(&ve->oui, oui, sizeof(oui));
	ve->oui_type = WFA_OUI_TYPE;

	attr = (struct qos_mgmt_attr_usr *)ve->data;

	switch (in->attribute_id)
	{
		case QOS_MGMT_ATTR_ID_PORT_RANGE:
		{
			BUF_PUT_BE16(attr->port_range.start_port,
			             in->port_range.start_port);
			BUF_PUT_BE16(attr->port_range.end_port, in->port_range.end_port);
			break;
		}
		case QOS_MGMT_ATTR_ID_DSCP_POLICY:
		{
			attr->dscp_policy.dscp_policy_id = in->dscp_policy.dscp_policy_id;
			attr->dscp_policy.request_type = in->dscp_policy.request_type;
			attr->dscp_policy.dscp = in->dscp_policy.dscp;
			break;
		}
		case QOS_MGMT_ATTR_ID_TCLAS:
		{
			cntlr_gen_qos_tclas_frame_classifier_elem(
				(struct tclas_frame_classifier *)&attr->tclas,
				&in->tclas.classifier);
			break;
		}
		case QOS_MGMT_ATTR_ID_DOMAIN_NAME:
		{
			memcpy(&attr->domain_name,
				   in->domain_name.domain_name,
				   in->domain_name.domain_name_len);
			break;
		}
		default:
		{
			break;
		}
	}

	return data;
}

#endif /* EASYMESH_VERSION >= 4 */
#endif /* EASYMESH_VERSION >= 3 */
